/*
*	SmallBASIC RTL - STANDARD COMMANDS
*
*	2000-05-27, Nicholas Christopoulos
*
*	This program is distributed under the terms of the GPL v2.0 or later
*	Download the GNU Public License (GPL) from www.gnu.org
*/

#define ERROR_CHECK_LEVEL 2
#include "sys.h"
#include "str.h"
#include "kw.h"
#include "panic.h"
#include "var.h"
#include "device.h"
#include "blib.h"
#include "pproc.h"
#include "smbas.h"

//
//	LET v[(x)] = any
//	CONST v[(x)] = any
//
void	cmd_let(int allowConst)
{
	var_t	*var_p = NULL;
	var_t	arg;
	int		nullarr = 0;

	if	( code_isnullarray_nc(0) )	{
		var_p = tvar[code_getnext16()];
		nullarr = 1;
		}
	else
		var_p = code_getvarptr();

	if	( !prog_error )	{
		if	( var_p->const_flag && !allowConst )	{
			rt_raise("CANNOT USE CONSTANT");
			return;		
			}

		if	( prog_source[prog_ip] == kwTYPE_CMPOPR && prog_source[prog_ip+1] == '=' )	
			code_skipnext16();

		if	( code_isnullarray() )	
			nullarr |= 2;

		if	( nullarr )	{
			if	( nullarr == 3 || nullarr == 2 )	{	// source is array
				code_skipnext();
				v_set(var_p, tvar[code_getnext16()]);
				var_p->const_flag = allowConst;
				}
			else if ( nullarr == 1 )	{	// dest is array but source isnt
				v_init(&arg);
				eval(&arg);

				if	( arg.type == V_ARRAY )	{
					v_set(var_p, &arg);
					v_free(&arg);
					var_p->const_flag = allowConst;
					}
				else	{
					v_free(&arg);
					rt_raise("LET: WRONG ASSIGNMENT");
					}
				}
			}
		else	{
			v_init(&arg);
			eval(&arg);
			v_set(var_p, &arg);
			v_free(&arg);
			var_p->const_flag = allowConst;
			}
		}
}

//
//	DIM	var([lower TO] uppper [, ...])
//
void	cmd_dim()
{
	int		exitf = 0;

	do	{
		var_t	*var_p = NULL;
		var_t	arg, array;
		int		i, size;
		byte	code;

		code = code_peek();
		if	( code == kwTYPE_LINE || code == kwTYPE_EOC )
			exitf = 1;
		else	{
			array.maxdim = 0;
			if	( code_peek() == kwTYPE_SEP )	{
				code_skipnext();
				if	( code_getnext() != ',' )
					err_syntax();
				}
			if	( code_peek() == kwTYPE_VAR )	{
				code_skipnext();
				var_p = tvar[code_getnext16()];
				if	( code_peek() == kwTYPE_LEVEL_BEGIN )	{
					code_skipnext();

					do	{
						eval(&arg);
						if	( prog_error )		return;

						if	( code_peek() == kwTO )	{
							array.lbound[array.maxdim] = v_getint(&arg);
							code_skipnext();
							eval(&arg);
							if	( prog_error )		return;

							array.ubound[array.maxdim] = v_getint(&arg);
							}
						else	{
							array.lbound[array.maxdim] = 0;
							array.ubound[array.maxdim] = v_getint(&arg);
							}
						v_free(&arg);
						array.maxdim ++;

						// skip separator
						if	( code_peek() == kwTYPE_SEP )	{
							code_skipnext();
							if	( code_getnext() != ',' )
								err_syntax();
							}

						} while ( code_peek() != kwTYPE_LEVEL_END );

					code_skipnext();	// ')', level
					}
				else	
					err_syntax();
				}
			else
				err_syntax();

			//
			//	run...
			//
			if	( !prog_error )	{
				size = 1;
				for ( i = 0; i < array.maxdim; i ++ )	
					size = size * (ABS(array.ubound[i] - array.lbound[i]) + 1);

				v_free(var_p);
				var_p->type = V_ARRAY;
				var_p->size = size;
				var_p->maxdim = array.maxdim;
				for ( i = 0; i < array.maxdim; i ++ )	{
					var_p->lbound[i] = array.lbound[i];
					var_p->ubound[i] = array.ubound[i];
					}
				var_p->ptr = tmp_alloc(size * sizeof(var_t));
				for ( i = 0; i < size; i ++ )
					v_init(v_getelemptr(var_p, i));
				}
			else
				exitf = 1;
			}
		} while ( !exitf && !prog_error );
}

//
//	ERASE var1 [, var2[, ...]]
//
void	cmd_erase()
{
	byte	code;
	var_t	*var_p;

	do	{
		if	( prog_error )
			break;

		code = code_peek();
		
		if ( code == kwTYPE_VAR )	{
			code_skipnext();
			var_p = tvar[code_getnext16()];
			}
		else	{
			err_typemismatch();
			break;
			}
		v_free(var_p);
		v_init(var_p);

		// next
		code = code_peek();
		if	( code == kwTYPE_SEP )	
			par_getcomma();	
		else
			break;

		} while ( 1 );
}

//
//	PRINT ...
//
void	cmd_print()
{
	byte	code, last_op = 0, exitf = 0;
	var_t	var;

	do	{
		code = code_peek();
		switch ( code )	{
		case	kwTYPE_LINE:
		case	kwTYPE_EOC:
			exitf = 1;
			break;
		case	kwTYPE_SEP:
			code_skipnext();
			last_op = code_getnext();
			if	( last_op == ',' )
				dev_print("\t");
			break;
		default:
			last_op = 0;
			v_init(&var);
			eval(&var);
			if	( !prog_error )
				print_var(&var);
			v_free(&var);
			};

		if	( prog_error )
			return;
		} while ( exitf == 0 );

	if	( last_op == 0 )
		dev_print("\n");
}

void	logprint_var(var_t *var)
{
	var_t	*vp;
	int		i;

	switch ( var->type )	{
	case	V_STR:
		lwrite(var->ptr);
		break;
	case	V_INT:
		lprintf("%d", var->i);
		break;
	case	V_NUM:
		lprintf("%f", var->n);
		break;
	case	V_ARRAY:	
		lprintf("\nArray of =%d elements\n", var->size+1);
		for ( i = 0; i < var->size; i ++ )	{
			vp = (var_t *) (var->ptr + (sizeof(var_t) * i));
			logprint_var(vp);
			}
		break;
		}
}

void	cmd_logprint()
{
	byte	code, last_op = 0, exitf = 0;
	var_t	var;

	do	{
		code = code_peek();
		switch ( code )	{
		case	kwTYPE_LINE:
		case	kwTYPE_EOC:
			exitf = 1;
			break;
		case	kwTYPE_SEP:
			code_skipnext();
			last_op = code_getnext();
			if	( last_op == ',' )
				lprintf("\t");
			break;
		default:
			last_op = 0;
			v_init(&var);
			eval(&var);
			if	( !prog_error )	
				logprint_var(&var);
			v_free(&var);
			};

		if	( prog_error )
			return;
		} while ( exitf == 0 );

	if	( last_op == 0 )
		lprintf("\n");
}

//
//	INPUT [str {,|;}] var [, var2 [, ...]]
//
void	cmd_input()
{
	byte		code, sep;
	var_t		arg;
	var_p_t		varlist[16];
	char_p_t	datalist[16];
	int			i, count, count_commas;
	char		*p, *ps, *np;
	char		prompt[256];
	char 		tmpsb[256];
	int			type;
	long		lv;
	double		dv;
	
	code = code_peek();
	strcpy(prompt, "");

	/* prompt */
	if	( code == kwTYPE_STR )	{
		v_init(&arg);
		eval(&arg);
		strcat(prompt, arg.ptr);
		v_free(&arg);

		code = code_getnext();
		if	( code == kwTYPE_SEP )	{
			sep = code_getnext();
			if	( sep == ';' )
				strcat(prompt, "? ");
			}
		else	{
			err_syntax();
			return;
			}

		}
	else // no prompt!
		strcat(prompt, "? ");

	/* count & check variables */
	count = 0;
	do	{
		code = code_peek();
		if	( code != kwTYPE_VAR )	{
			err_syntax();
			return;
			}
		code_skipnext();

		// check 	
		if	( count == 15 )	{
			rt_raise("INPUT: TOO MANY VARS");
			return;
			}

		// store variable's pointer
		varlist[count] = code_getvarptr();
		if	( prog_error )	return;
		if	( varlist[count]->const_flag )	{
			rt_raise("INPUT: CANNOT USE CONSTANT");
			return;
			}
		v_free(varlist[count]);
		count ++;

		// next
		code = code_peek();
		if	( code != kwTYPE_SEP )
			break;
		else	
			code_skipnext16();
		} while ( 1 );

	/* read the string */
	do	{
		dev_print(prompt);
		dev_gets(tmpsb, 255);
		if	( prog_error )	return;		// break or something can be happened
		
		// check user's input string
		if	( count > 1 )	{
			count_commas = 0;
			p = tmpsb;
			while ( *p )	{
				if	( *p == ',' )	{
					if ( count_commas == count )	{
						count_commas = -1;
						break;
						}

					// store ptr
					datalist[count_commas] = p;
					count_commas ++;
					}
				p ++;
				}
			}
		else
			count_commas = 0;	// only one variable is rq

		if	( count != (count_commas+1) )	// error number of input variables
			dev_print("* REDO FROM START *\n");

		} while ( count != (count_commas+1) );

	/* assign, variables */
	for ( i = 0; i < count; i ++ )	{
		// ps = string data start
		if	( i > 0 )	{
			ps = datalist[i-1];
			ps ++;
			}
		else
			ps = tmpsb;

		// close the string
		if	( i < count_commas )	{
			p = datalist[i];
			*p = '\0';
			}

		// 
		np = get_numexpr(ps, prompt, &type, &lv, &dv);

		if	( type == 1 && *np == '\0' )	{
			varlist[i]->type = V_INT;
			varlist[i]->i = lv;
			}
		else if	( type == 2  && *np == '\0' )	{
			varlist[i]->type = V_NUM;
			varlist[i]->n = dv;
			}
		else	
			v_createstr(varlist[i], ps);
		}
}

//
//	ON x GOTO|GOSUB ...
//
void	cmd_on_go()
{
	word	next_ip, expr_ip, table_ip, dest_ip;
	byte	command, count;
	var_t	var;
	word	index;
	stknode_t	node;

	next_ip = code_getnext16();
	code_skipnext16();
	command  = code_getnext();
	count    = code_getnext();
	table_ip = prog_ip;
	expr_ip  = prog_ip + (count << 1);

	v_init(&var);
	prog_ip = expr_ip;
	eval(&var);
	index = (v_igetval(&var) - 1);
	if	( index >= count )
		rt_raise("ON x %s: OUT OF RANGE", (command==kwGOTO)?"GOTO":"GOSUB");
	
	memcpy(&dest_ip, prog_source+table_ip+(index<<1), 2);
	switch ( command )	{
	case	kwGOTO:
		code_jump(dest_ip);
		break;
	case	kwGOSUB:
		node.type = kwGOSUB;
		node.x.vgosub.ret_ip = next_ip;
		code_jump(dest_ip);
		code_push(&node);
		break;
	default:
		rt_raise("ON x: INTERNAL ERROR");
		}
}

//
//	GOSUB label
//
void	cmd_gosub()
{
	stknode_t	node;
	word		goto_label;

	goto_label = code_getnext16();
	node.type = kwGOSUB;
	node.x.vgosub.ret_ip = prog_ip;
	code_jump_label(goto_label);
	code_push(&node);
}

//
//	CALL UDP/F address
//
//	Stack:
//	[p1]
//	...
//	[pN]
//	[vcall]
//
//	p1...pN will be removed by cmd_param()
//
void	cmd_udp(int cmd)
{
	stknode_t	ncall, param;
	word		goto_addr, pcount = 0, rvid;
	var_t		*var_p, *arg = NULL;
	byte		ready, code;
	word		ofs;

	goto_addr = code_getnext16();		// jump to
	rvid      = code_getnext16();		// return-variable ID

	if	( code_peek() == kwTYPE_LEVEL_BEGIN )	{
		code_skipnext();		// kwTYPE_LEVEL_BEGIN

		ready = 0;
		do	{
			code = code_peek();
			switch ( code )	{
			case	kwTYPE_EOC:
				code_skipnext();
				break;
			case	kwTYPE_SEP:			// separator 
				code_skipnext16();
				break;
			case	kwTYPE_LEVEL_END:	// ) -- end of parameters
				ready = 1;
				break;
			case	kwTYPE_VAR:		// variable
				ofs = prog_ip;		// store IP

				// check if its a whole array i.e. MAX(v())
				if	( code_isnullarray() )	{	
					code_skipnext();

					// push parameter
					param.type = kwTYPE_VAR;
					param.x.param.res = tvar[code_getnext16()];
					param.x.param.vcheck = 0x83;	// NULLARRAY - BYVAL|BYREF
					code_push(&param);
					pcount ++;
					break;
					}

				code_skipnext();
				var_p = code_getvarptr();
			
				// check if it has no more args (sep|rpar)
				code = code_peek();
				if	( code == kwTYPE_EOC )	{
					code_skipnext();
					code = code_peek();
					}

				if ( code == kwTYPE_SEP || code == kwTYPE_LEVEL_END )	{
					// push parameter
					param.type = kwTYPE_VAR;
					param.x.param.res = var_p;
					param.x.param.vcheck = 3;	// BYVAL|BYREF
					code_push(&param);
					pcount ++;
					break;
					}

				prog_ip = ofs;	// restore IP
				// no 'break' here
			default:
				// default --- expression (BYVAL ONLY)
				arg = (var_t *) tmp_alloc(sizeof(var_t));
				v_init(arg);

				eval(arg);
				if	( !prog_error )	{
					// push parameter
					param.type = kwTYPE_VAR;
					param.x.param.res = arg;	
					if	( arg->type == V_ARRAY )
						param.x.param.vcheck = 0x81;	// ARRAY - BYVAL ONLY
					else
						param.x.param.vcheck = 1;	// BYVAL ONLY
					code_push(&param);
					pcount ++;
					}
				else	{
					v_free(arg);
					return;
					}

				}

			} while ( !ready );

		///
		code_skipnext();		// kwTYPE_LEVEL_END
		}

	///
	ncall.type = cmd;
	ncall.x.vcall.pcount = pcount;
	ncall.x.vcall.ret_ip = prog_ip;
	ncall.x.vcall.rvid   = rvid;
	if	( rvid != 0xFFFF )	{
		ncall.x.vcall.retvar = tvar[rvid];	// store previous data of RVID
		tvar[rvid] = v_new();
		}

	code_push(&ncall);
	prog_ip = goto_addr;
}

//
//	RETURN FROM UDP/F
//
void	cmd_udpret()
{
	stknode_t	node, rval;
	
	code_pop(&node);
	while ( (node.type != kwPROC) && (node.type != kwFUNC) )	{ 

		if	( node.type == kwTYPE_CRVAR )	{ 				// local variable - cleanup
			v_free(tvar[node.x.vdvar.vid]);					// free local variable data
			tmp_free(tvar[node.x.vdvar.vid]);
			tvar[node.x.vdvar.vid] = node.x.vdvar.vptr;		// restore ptr
			}
		else if	( node.type == kwBYREF )	{ 			// variable reference
			tvar[node.x.vdvar.vid] = node.x.vdvar.vptr;		// restore ptr
			}

		// next
		code_pop(&node); 
		if ( prog_error )
			return; 
		}

	if ( (node.type != kwPROC) && (node.type != kwFUNC) )	{
		err_syntax();
		dump_stack();
		}
	else	{
		// restore return value
		if	( node.x.vcall.rvid != 0xFFFF )	{

			// store value to stack
			rval.type = kwTYPE_RET;
			rval.x.vdvar.vptr = tvar[node.x.vcall.rvid];
			code_push(&rval);

			tvar[node.x.vcall.rvid] = node.x.vcall.retvar;	// restore ptr
			}
		}

	prog_ip = node.x.vcall.ret_ip;
}

//
//	CREATE DYNAMIC VARIABLES
//
void	cmd_crvar()
{
	int			i, count;
	word		vid;
	stknode_t	node;

	count = code_getnext();
	for ( i = 0; i < count; i ++ )	{
		vid   = code_getnext16();

		// store previous variable-ptr to stack
		node.type = kwTYPE_CRVAR;
		node.x.vdvar.vid = vid;
		node.x.vdvar.vptr = tvar[vid];
		code_push(&node);
		
		// create a new one with the same ID
		tvar[vid] = (var_t *) tmp_alloc(sizeof(var_t));
		v_init(tvar[vid]);
		}
}

//
//	UDP/F PARAMETERS
//
void	cmd_param()
{
	int		i, pcount;
	int		vid;
	byte	vattr;
	stknode_t	ncall, *param, node;

	code_pop(&ncall);
	if ( (ncall.type != kwPROC) && (ncall.type != kwFUNC) )	{ 
		err_stackmess();
		return;
		}
	pcount = code_getnext();

	if	( pcount != ncall.x.vcall.pcount )	{
		err_parm_num();
		return;
		}

	if	( pcount )	{
		// get parameters
		param = (stknode_t *) tmp_alloc(sizeof(stknode_t) * pcount);
		for ( i = pcount-1; i > -1; i -- )	
			code_pop(&param[i]);

		// push call's pars (again)
		code_push(&ncall);

		//
		for ( i = 0; i < pcount; i ++ )	{
			vattr = code_getnext();
			vid   = code_getnext16();

			if	( (vattr & 0x80) == 0 )	{		// BY VALUE
				// store previous variable to stack
				node.type = kwTYPE_CRVAR;
				node.x.vdvar.vid = vid;
				node.x.vdvar.vptr = tvar[vid];
				code_push(&node);

				// assign
				if	( param[i].x.param.vcheck == 1 )	{	
					if	( vattr & 1 )	{
						code_pop(&node);
						rt_raise("PARAM %d IS AN ARRAY", i+1);
						}
					else
						tvar[vid] = param[i].x.param.res;
					}
				else if ( param[i].x.param.vcheck == 0x81 )	{
					if	( (vattr & 1) == 0 )	{
						code_pop(&node);
						rt_raise("PARAM %d IS NOT AN ARRAY", i+1);
						}
					else
						tvar[vid] = param[i].x.param.res;
					}

				else if ( param[i].x.param.vcheck & 0x80 )	{
					if	( (vattr & 1) == 0 )	{
						code_pop(&node);
						rt_raise("PARAM %d IS NOT AN ARRAY", i+1);
						}
					else
						tvar[vid] = v_clone(param[i].x.param.res);
					}
				else	{
					if	( vattr & 1 )	{
						code_pop(&node);
						rt_raise("PARAM %d IS AN ARRAY", i+1);
						}
					else
						tvar[vid] = v_clone(param[i].x.param.res);
					}
				}
			else	{					// BY REFERENCE
				if	( (param[i].x.param.vcheck == 1) || (param[i].x.param.vcheck == 0x81) )	{
					err_parm_byref(i);
					break;
					}
				else	{
					// store previous variable to stack
					node.type = kwBYREF;
					node.x.vdvar.vid = vid;
					node.x.vdvar.vptr = tvar[vid];
					code_push(&node);

					if ( param[i].x.param.vcheck & 0x80 )	{
						if	( (vattr & 1) == 0 )	{
							code_pop(&node);
							rt_raise("PARAM %d IS NOT AN ARRAY", i+1);
							}
						else
							tvar[vid] = param[i].x.param.res;
						}
					else	{
						if	( vattr & 1 )	{
							code_pop(&node);
							rt_raise("PARAM %d IS AN ARRAY", i+1);
							}
						else
							tvar[vid] = param[i].x.param.res;
						}

					}
				}
			}

		tmp_free(param);
		}
	else	{
		// push call's pars (again)
		code_push(&ncall);
		}
}

//
//	EXIT [FOR|LOOP|FUNC|PROC]
//
void	cmd_exit()
{
	stknode_t	node;
	int			ready = 0;
	word		addr = 0xFFFF;
	byte		code;

	code = code_peek();
	if	( !(code == kwLOOP || code == kwFOR || 
		    code == kwPROC || code == kwFUNC) )
		code = 0;
	else
		code_skipnext();

	do	{
		code_pop(&node);
		if	( prog_error )
			return;
		
		switch ( node.type )	{
		case	kwIF:
			break;
		case	kwGOSUB:
			if	( code == 0 )	{
				addr = node.x.vgosub.ret_ip;
				ready = 1;
				}
			break;
		case	kwFOR:
			if	( code == 0 || code == kwFOR )	{
				addr = node.exit_ip;
				ready = 1;
				}
			break;
		case	kwWHILE:
			if	( code == 0 || code == kwLOOP )	{
				addr = node.exit_ip;
				ready = 1;
				}
			break;
		case	kwREPEAT:
			if	( code == 0 || code == kwLOOP )	{
				// address of UNTIL
				prog_ip = node.exit_ip+1;
				addr = code_getnext16();
				ready = 1;
				}
			break;
		case	kwPROC:
		case	kwFUNC:
		case	kwTYPE_CRVAR:
		case	kwBYREF:
		case	kwTYPE_PARAM:
			if	( code == 0 || code == kwPROC || code == kwFUNC )	{
				code_push(&node);
				cmd_udpret();
				addr = 0xFFFF;
				ready = 1;
				}
			else	{
				if	( code == kwFOR )
					rt_raise("EXIT FOR: NO 'FOR' INSIDE SUB/FUNC");
				else
					rt_raise("EXIT LOOP: NO 'LOOP' INSIDE SUB/FUNC");
				}
			break;
			};
		} while ( ready == 0 );

	if	( addr != 0xFFFF )
		code_jump(addr);
}

//
//	RETURN
//
void	cmd_return()
{
	stknode_t	node;
	
	// get return-address and remove any other item (sub items) from stack
	code_pop(&node);

	// 'GOTO' SHIT
	while ( node.type != kwGOSUB )	{ code_pop(&node); if ( prog_error ) return; }

	if ( node.type != kwGOSUB )	{
		err_syntax();
		dump_stack();
		}

	code_jump(node.x.vgosub.ret_ip);
}

//
//	IF expr [THEN]
//
void	cmd_if()
{
	word	true_ip, false_ip;
	var_t	var;
	stknode_t	node;

	true_ip = code_getnext16();
	false_ip = code_getnext16();

	// expression
	v_init(&var);
	eval(&var);
	node.type = kwIF;
	node.x.vif.lcond = v_is_nonzero(&var);
	code_jump( (node.x.vif.lcond) ? true_ip : false_ip);
	v_free(&var);

	code_push(&node);
}

//
//	ELSE
//
void	cmd_else()
{
	word	true_ip, false_ip;
	stknode_t	node;

	true_ip = code_getnext16();
	false_ip = code_getnext16();

	code_pop(&node);

	// 'GOTO' SHIT
	while ( node.type != kwIF )	{ code_pop(&node); if ( prog_error ) return; }

	if	( node.type != kwIF )	{
		err_syntax();
		dump_stack();
		return;
		}
	code_push(&node);
			
	code_jump((!node.x.vif.lcond) ? true_ip : false_ip);
}

//
//	ELIF
//
void	cmd_elif()
{
	word		true_ip, false_ip;
	var_t		var;
	stknode_t	node;

	true_ip = code_getnext16();
	false_ip = code_getnext16();

	// else cond
	code_pop(&node);

	// 'GOTO' SHIT
	while ( node.type != kwIF )	{ code_pop(&node); if ( prog_error ) return; }

	if	( node.type != kwIF )	{
		err_syntax();
		dump_stack();
		return;
		}

	if	( !node.x.vif.lcond )			{
		// previous IF failed

		// expression
		v_init(&var);
		eval(&var);
		node.x.vif.lcond = v_is_nonzero(&var);
		code_jump( (node.x.vif.lcond) ? true_ip : false_ip);
		v_free(&var);

		code_push(&node);
		}
	else	{
		// previous IF succeded
		code_push(&node);
		code_jump(false_ip);
		}
}

//
//	FOR var = expr TO expr [STEP expr]
//
void	cmd_for()
{
	byte	code;
	word	true_ip, false_ip;
	stknode_t	node;
	var_t	var, varstep, *var_p;

	true_ip = code_getnext16();
	false_ip = code_getnext16();

	code = code_peek();
	if	( code != kwTYPE_VAR )	{
		err_syntax();
		return;
		}
	node.type = kwFOR;
	node.exit_ip = false_ip + 5;
	node.x.vfor.jump_ip = true_ip;

	//
	//	get FOR-variable
	//
	code_skipnext();
	var_p = code_getvarptr();
	if	( prog_error )	return;
	node.x.vfor.var_ptr = var_p;
	v_free(var_p);

	v_init(&var);
	v_init(&varstep);

	// get the first expression
	eval(&var);
	if	( !prog_error && (var.type == V_NUM || var.type == V_INT) )	{
		//
		//	assign FOR-variable
		//
		v_set(var_p, &var);

		code = code_getnext();
		if	( code == kwTO )	{
			//
			//	get TO-expression
			//
			node.x.vfor.to_expr_ip = prog_ip;
			v_init(&var);
			eval(&var);	

			if	( !prog_error && (var.type == V_NUM || var.type == V_INT) )	{
				//
				//	step
				//
				code = code_peek();
				if	( code == kwSTEP )	{
					code_skipnext();
					node.x.vfor.step_expr_ip = prog_ip;
					eval(&varstep);	
					if	( !(varstep.type == V_NUM || varstep.type == V_INT) )	{
						if	( !prog_error )
							err_syntax();	
						}
					}
				else	{
					node.x.vfor.step_expr_ip = 0xFFFF;
					varstep.type = V_INT;
					varstep.i = 1;
					} // STEP kw
				}  
			else {	// str for TO
				if	( !prog_error )
					err_syntax();	
				}
			}
		else	// TO keyword
			err_syntax();
		}

	//
	//	run
	//	
	if	( !prog_error )	{
		if	( v_sign(&varstep) < 0 )	{
			code_jump( (v_compare(var_p, &var) >= 0) ? true_ip : false_ip);
			}
		else {
			if	( v_compare(var_p, &var) <= 0 )	
				code_jump(true_ip);
			else	
				code_jump(false_ip);
			}

		code_push(&node);
		}

	// clean up
	v_free(&var);
	v_free(&varstep);
}

//
//	WHILE expr
//
void	cmd_while()
{
	word	true_ip, false_ip;
	var_t	var;
	stknode_t	node;

	true_ip = code_getnext16();
	false_ip = code_getnext16();

	// expression
	v_init(&var);
	eval(&var);
	if	( v_sign(&var) )
		code_jump(true_ip);
	else
		code_jump(false_ip+5);
	v_free(&var);

	node.type = kwWHILE;
	node.exit_ip = false_ip+5;
	code_push(&node);	// GOTO's SHIT
}

//
//	WEND
//
void	cmd_wend()
{
	stknode_t	node;
	word	/* next_ip, */ jump_ip;

	/* next_ip = */ code_skipnext16();
	jump_ip = code_getnext16();
	code_jump(jump_ip);
	code_pop(&node);			// GOTO's SHIT
}

//
//	UNTIL expr
//
void	cmd_until()
{
	word	/* next_ip, */ jump_ip;
	var_t	var;
	stknode_t	node;

	/* next_ip = */ code_skipnext16();
	jump_ip = code_getnext16();

	// expression
	v_init(&var);
	eval(&var);
	if	( !v_sign(&var) )
		code_jump(jump_ip);
	v_free(&var);
	code_pop(&node);			// GOTO's SHIT
}

//
//	NEXT
//
void	cmd_next()
{
	word	next_ip, jump_ip;
	var_t	var_to, var_step, *var_p;
	int		check = 0;
	stknode_t	node;

	next_ip = code_getnext16();
	jump_ip = code_getnext16();

	code_pop(&node);

	// 'GOTO' SHIT
	while ( node.type != kwFOR )	{ code_pop(&node); if ( prog_error ) return; }

	if	( node.type != kwFOR )	{
		err_syntax();
		dump_stack();
		return;
		}
	jump_ip = node.x.vfor.jump_ip;

	//
	var_p = node.x.vfor.var_ptr;

	v_init(&var_to);
	v_init(&var_step);

	// get last val
	prog_ip = node.x.vfor.to_expr_ip;
	eval(&var_to);
	if	( !prog_error && (var_to.type == V_INT || var_to.type == V_NUM) )	{
		// get step val
		if	( node.x.vfor.step_expr_ip == 0xFFFF )	{
			var_step.type = V_INT;
			var_step.i = 1;
			}
		else	{
			prog_ip = node.x.vfor.step_expr_ip;
			eval(&var_step);
			}

		if	( !prog_error && (var_step.type == V_INT || var_step.type == V_NUM) )	{
			v_inc(var_p, &var_step);
			if	( v_sign(&var_step) < 0 )	
				check = (v_compare(var_p, &var_to) >= 0);
			else 
				check = (v_compare(var_p, &var_to) <= 0);
			} //
		else	{
			if	( !prog_error )
				err_syntax();
			}
		}
	else	{
		if	( !prog_error )
			err_syntax();
		}

	//
	//	RUN
	//
	if	( !prog_error )	{
		if	( check )	{
			code_push(&node);
			code_jump(jump_ip);	
			}
		else
			code_jump(next_ip);	
		}

	// clean up
	v_free(&var_to);
	v_free(&var_step);
}

//
//	READ var [, var [, ...]]
//
void	cmd_read()
{
	byte	code, exitf = 0;
	var_t	*vp = NULL;

	if	( data_ip == 0xFFFF )	{
		rt_raise("READ: USE RESTORE [label]");
		return;
		}

	do	{
		code = code_peek();
		switch ( code )	{
		case	kwTYPE_LINE:
		case	kwTYPE_EOC:
			exitf = 1;
			break;
		case	kwTYPE_SEP:
			code_skipnext16();	// get 2 bytes
			break;
		default:
			code_skipnext();
			vp = code_getvarptr();
			if	( prog_error )	return;

			//////
			if	( !prog_error )	{
				v_free(vp);

				if	( data_ip >= prog_length )	{
					rt_raise("READ: OUT OF RANGE(DATA)");
					return;
					}

				switch ( prog_source[data_ip] )	{
				case	kwTYPE_EOC:	// null data
					vp->type = V_INT;
					vp->i = 0;
					break;
				case	kwTYPE_INT:
					data_ip ++;

					vp->type = V_INT;
					memcpy(&vp->i, prog_source+data_ip, 4);
					data_ip += 4;
					break;
				case	kwTYPE_NUM:
					data_ip ++;

					vp->type = V_NUM;
					memcpy(&vp->n, &prog_source[data_ip], 8);
					data_ip += 8;
					break;
				case	kwTYPE_STR:
					{
					word	len;

					data_ip ++;

					vp->type = V_STR;
					memcpy(&len, prog_source+data_ip, 2);
					data_ip += 2;

					vp->ptr = tmp_alloc(len+1);
					memcpy(vp->ptr, prog_source+data_ip, len);
					*((char *) (vp->ptr+len)) = '\0';
					vp->size = len;
					data_ip += len;
					}
					break;
				default:
					rt_raise("READ: OUT OF RANGE(DATA) IDX=%d", data_ip);
					return;
					}

				if	( prog_source[data_ip] == kwTYPE_EOC )	
					data_ip ++;

				}
			};
		} while ( !exitf ); 
}

//
//	DATA ...
//
void	cmd_data()
{
	rt_raise("DATA: CANNOT EXECUTE DATA");
}

//
//	RESTORE label
//
void	cmd_restore()
{
	data_ip = code_getnext16();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////	SYSTEM
////////////

//
//	RANDOMIZE [num]
//
void	cmd_randomize()
{
	byte	code;
	long	seed;

	code = code_peek();
	switch ( code )	{
	case	kwTYPE_LINE:
	case	kwTYPE_EOC:
		#if defined(_PalmOS)
		SysRandom(TimGetTicks());
		#else
		srand(clock());
		#endif
		break;
	default:
		seed = par_getint();
		if ( !prog_error )	{
			#if defined(_PalmOS)
			SysRandom(seed);
			#else
			srand(seed);
			#endif
			}
		};
}

//
//	DELAY
//
void	cmd_delay()
{
	dword	ms;

	ms = par_getint();		if	( prog_error )	return;
	dev_delay(ms);
}

//
//	AT x,y
//
void	cmd_at()
{
	int		x, y;

	x = par_getint();		if	( prog_error )	return;
	    par_getcomma();		if	( prog_error )	return;
	y = par_getint();		if	( prog_error )	return;
	dev_setxy(x,y);
}

//
//	LOCATE y,x
//
void	cmd_locate()
{
	int		x, y;

	y = par_getint();		if	( prog_error )	return;
	    par_getcomma();		if	( prog_error )	return;
	x = par_getint();		if	( prog_error )	return;
	#if defined(_PalmOS)
	dev_setxy(x*8,y*11);
	#else
	dev_setxy(x*8,y*16);
	#endif
}

//
//	PAUSE [secs]
//
void	cmd_pause()
{
	int		x = 0, evc;
	byte	code;
	long	start, now, tps;

	code = code_peek();
	if	( code == kwTYPE_VAR ) {
		x = par_getint();	
		if	( prog_error )	return;
		}
	else if ( code == kwTYPE_INT )	{
		code_skipnext();
		x = code_getnext32();
		}
	else if ( code == kwTYPE_NUM )	{
		code_skipnext();
		x = code_getnext64f();
		}

	if	( x == 0 )	{
		while ( dev_kbhit() == 0 )	{
			switch ( dev_events(0) )	{
			case	0:	// no event
				break;
			case	-2:	// break
				brun_break();
			case	-1:	// break
				return;
				}
			}

		dev_getch();
		}
	else	{
		#if defined(_PalmOS)
		start = TimGetTicks();
		tps = SysTicksPerSecond();
		#else
		start = clock();
		tps = CLOCKS_PER_SEC;
		#endif
		while ( !dev_kbhit() )	{
			switch ( (evc = dev_events(0)) )	{
			case	0:	// no event
				break;
			case	-2:	// break
				brun_break();
			case	-1:	// break
				return;
				}
			if	( evc )	{
				if	( dev_kbhit() )
					dev_getch();
				break;
				}

			#if defined(_PalmOS)
			now = TimGetTicks();
			#else
			now = clock();
			#endif
			
			if	( now > start + tps * x )
				break;
			}

		dev_getch();
		}
}

//
//	COLOR fg[,text_bg]
//
void	cmd_color()
{
	int		fg, bg = -1;

	fg = par_getint();		if	( prog_error )	return;
	if	( code_peek() == kwTYPE_SEP )	{
	    par_getcomma();			if	( prog_error )	return;
		bg = par_getint();		if	( prog_error )	return;
		}

	dev_settextcolor(fg, bg);
}

//
//	SPLIT string, delimiters, array()
//
void	cmd_wsplit()
{
	int		quotes, len, count, lc = 0;
	char	*p, *ps, *new_text;
	var_t	str, del, *var_p, *elem_p;

	v_init(&str);
	v_init(&del);
	par_getstr(&str);	if	( prog_error )	return;
	par_getcomma();		if	( prog_error )	{ v_free(&str); return; }
	par_getstr(&del);	if	( prog_error )	{ v_free(&str); return; }
	par_getcomma();		if	( prog_error )	{ v_free(&str); v_free(&del); return; }
	if	( !code_isnullarray() )	{
		if	( code_peek() == kwTYPE_VAR )	{
			code_skipnext();
			var_p = tvar[code_getnext16()];
			v_free(var_p);

			var_p->type = V_ARRAY;
			var_p->ptr = tmp_alloc(sizeof(var_t));
			var_p->size = 1;
			var_p->lbound[0] = 0;
			var_p->ubound[0] = var_p->lbound[0] + 1;
			var_p->maxdim = 1;
			}
		else	{
			err_typemismatch();
			v_free(&str); v_free(&del); 
			return;
			}
		}
	else	{
		code_skipnext();
		var_p = tvar[code_getnext16()];
		}

	//
	//	reformat
	//
	p = (char *) str.ptr;
	new_text = tmp_alloc((len=(strlen(p)+4)));
	memset(new_text, 0, len);
	ps = new_text;
	quotes = 0;
	while ( *p )	{
		if	( !quotes )	{
			if ( *p == '\n' )	{
				*ps ++ = '\n';
				while ( *(p+1) == ' ' || *(p+1) == '\t' )
					p ++;
				}
			else if ( (*p == '\t' || *p == ' ') && lc != ' ' )
				*ps ++ = ' ';
			else if	( (*p > ' ') || (*p & 0x80) )	
				*ps ++ = *p;

			lc = *ps;
			}
		else
			*ps ++ = *p;

		if	( *p == '\"' )	
			quotes = !quotes;

		p ++;
		}
	*ps = '\0';	

	//
	//	scan
	//
	count = 0;
	ps = p = new_text;
	while ( *p )	{
		if	( strchr(del.ptr, *p) )	{
			*p = '\0';

			// add element (ps)
			if	( var_p->size <= count )		// resize array
				v_resize_array(var_p, count+8);

			elem_p = (var_t *) (var_p->ptr + (sizeof(var_t) * count));
			elem_p->type = V_STR;
			elem_p->ptr = tmp_alloc((len=(strlen(ps)+1)));
			strcpy(elem_p->ptr, ps);
			elem_p->size = len;

			count ++;

			// next word
			ps = p+1;
			}

		p ++;
		}

	if	( *ps )	{
		// add the last element (ps)
		if	( var_p->size <= count )		// resize array
			v_resize_array(var_p, count+8);

		elem_p = (var_t *) (var_p->ptr + (sizeof(var_t) * count));
		elem_p->type = V_STR;
		elem_p->ptr = tmp_alloc((len=(strlen(ps)+1)));
		strcpy(elem_p->ptr, ps);
		elem_p->size = len;

		count ++;
		}

	v_resize_array(var_p, count);	// final resize

	// cleanup
	v_free(&str);
	v_free(&del);
	tmp_free(new_text);
}

#ifdef BC_DEBUG
//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

void	cmd_bcdump()
{
	int		c, i, b, h, l, j;
	int		len, new_ip;
	int		a = 10; // Default dump size
	char	buf[10];

	b = code_getnext();
	while ( b == kwTYPE_EOC )	b = code_getnext();

	if ( b == kwTYPE_INT ) {
		a = code_getnext32();
	   	b = code_getnext();
		while ( b == kwTYPE_EOC )	b = code_getnext();
		}

	if ( b != kwTYPE_LINE) {
		dev_printf("** BCDUMP [n] **\n");
		return;
		}
	else
		dev_printf("== BCDUMP-LINE: %d\n", code_getnext16());

	for ( i = 0; i < a; i ++ ) {
		b = code_getnext();
		h = b >> 4;
		l = b & 0xF;

		dev_printf("%04d: %c%c %c ",
			prog_ip-1,
			"0123456789ABCDEF"[h],
			"0123456789ABCDEF"[l],
			(b>=32) ? b : 42 );

		switch ( b )	{
		case	kwTYPE_LINE:
			dev_printf("--LINE: %d", code_getnext16());
			break;
		case	kwTYPE_EOC:
			dev_printf("--EOC");
			break;
		default:
			for ( c = 0; keyword_table[c].name[0] != '\0'; c ++)	{
				if ( b == keyword_table[c].code )	{
					dev_printf("%s ", keyword_table[c].name);
					break;
					}
				}

			switch ( b )	{
			case	kwTYPE_LEVEL_BEGIN:
			case	kwTYPE_LEVEL_END:
				break;
			case	kwTYPE_SEP:
			case	kwTYPE_LOGOPR:
			case	kwTYPE_CMPOPR:
			case	kwTYPE_ADDOPR:
			case	kwTYPE_MULOPR:
			case	kwTYPE_POWOPR:
			case	kwTYPE_UNROPR:
				dev_printf(" SYM '%c'", code_getnext());
				break;
			case	kwFILEINPUT:
				dev_printf("FILE INPUT");
				break;
			case	kwFILEPRINT:
				dev_printf("FILE PRINT");
				break;
			case	kwONJMP:
				new_ip = code_getnext16();
				code_skipnext16();
				b = code_getnext();
				c = code_getnext();
				dev_printf("ON %s CNT %d\n", ((b == kwGOTO)?"GOTO":"GOSUB"), c);
				dev_printf("      NEXT IP %d\n", new_ip);
				for ( j = 0; j < c; j ++ )	
					dev_printf("       JMP IP %d\n", code_getnext16());
				dev_printf("ON %s EXPR:", ((b == kwGOTO)?"GOTO":"GOSUB"));
				break;
			case	kwGOTO:
				new_ip = code_getnext16();
				dev_printf("= JMP %d ", new_ip);
				c = code_getnext();
				dev_printf(" POPs %d", c);
				break;
			case	kwGOSUB:
				c = code_getnext16();
				dev_printf("ID %d ", c);
				new_ip = tlab[c].ip;
				dev_printf("= JMP %d ", new_ip);
				dev_printf("\n(J)ump, (C)ontinue?");
				do	{
					dev_gets(buf, 8);
					if	( buf[0] == 'j' )	{
						prog_ip = new_ip;
						break;
						}
					else if	( buf[0] == 'c' )
						break;
					} while ( 1 );

				break;
			case	kwPROC:
				dev_printf("=== PROCEDURE ===");
				break;
			case	kwFUNC:
				dev_printf("=== FUNCTION ===");
				break;
			case	kwTYPE_PARAM:
			case	kwTYPE_CRVAR:
				c = code_getnext();
				if	( b == kwTYPE_PARAM )
					dev_printf("PARAM = %d: ", c);
				else
					dev_printf("CRVAR = %d: ", c);
				
				for ( j = 0; j < c; j ++ )	{
					if	( b == kwTYPE_PARAM )	{
						if	( code_getnext() & 0x80 )
							dev_printf("&");
						}
					dev_printf("%d ", code_getnext16());
					}
				break;
			case	kwTYPE_RET:
				dev_printf("=== UDP/F RET ===");
				break;
			case	kwTYPE_CALL_UDP:
				dev_printf("CALL UDP %d", code_getnext16());
				code_skipnext16();
				break;
			case	kwTYPE_CALL_UDF:
				dev_printf("CALL UDF %d,", code_getnext16());
				dev_printf(" RVID: %d", code_getnext16());
				break;
			case	kwRESTORE:
			case	kwTYPE_LINE:
			case	kwLABEL:
			case	kwTYPE_VAR:
			case	kwLET:
				dev_printf("ID %d ", code_getnext16());
				break;
			case	kwTYPE_INT:
				dev_printf("VAL %d ", code_getnext32());
				break;
			case	kwTYPE_NUM:
				dev_printf("VAL %f ", code_getnext64f());
				break;
			case	kwTYPE_STR:
				len = code_getnext16();
				dev_printf("\"");
				for ( j = 0; j < len; j ++ )
					dev_printf("%c", prog_source[prog_ip+j]);
				dev_printf("\"");
				prog_ip += len;	// +1?
				break;
			case	kwIF:		case	kwFOR: 		case	kwWHILE:	case	kwREPEAT:
			case	kwELSE:		case	kwELIF:
			case	kwENDIF:	case	kwNEXT:		case	kwWEND:		case	kwUNTIL:
				dev_printf("IP1 %d IP2 %d ", code_getnext16(), code_getnext16());
				break;
				}
			}
		dev_printf("\n");
		if ( b == kwSTOP ) break;
	}
}
#endif

